
#include "haskell-curses.h"

module UI.Curses where

import UI.Curses.Internals (gen_decs)
import Control.Monad.State (MonadIO, liftIO)
import System.Cmd (system)
import Data.Char (ord, toUpper)
import Foreign.Marshal.Alloc (alloca, free)
import Foreign.Marshal.Array (newArray0)
import Foreign.Storable (peek)
import Foreign.Ptr (Ptr, nullPtr)
import Foreign.C.Types (CShort, CInt, CWchar)
import Foreign.C.String (CString, newCString)
import Data.Word (#type bool
                 ,#type chtype
                 ,#type attr_t)
import Language.Haskell.TH

data Window = Window
type PWindow = Ptr Window
type CBool = #type bool
type ChType = #type chtype
type Attr = #type attr_t

cERR :: CInt
cERR = #const ERR

cKEY_CODE_YES, cKEY_BREAK, cKEY_DOWN, cKEY_UP, cKEY_LEFT, cKEY_RIGHT, cKEY_HOME, cKEY_BACKSPACE, cKEY_F0, {- cKEY_F(cn), -} cKEY_DL, cKEY_IL, cKEY_DC, cKEY_IC, cKEY_EIC, cKEY_CLEAR, cKEY_EOS, cKEY_EOL, cKEY_SF, cKEY_SR, cKEY_NPAGE, cKEY_PPAGE, cKEY_STAB, cKEY_CTAB, cKEY_CATAB, cKEY_ENTER, cKEY_SRESET, cKEY_RESET, cKEY_PRINT, cKEY_LL, cKEY_A1, cKEY_A3, cKEY_B2, cKEY_C1, cKEY_C3 :: CInt
cKEY_CODE_YES = #const KEY_CODE_YES
cKEY_BREAK = #const KEY_BREAK
cKEY_DOWN = #const KEY_DOWN
cKEY_UP = #const KEY_UP
cKEY_LEFT = #const KEY_LEFT
cKEY_RIGHT = #const KEY_RIGHT
cKEY_HOME = #const KEY_HOME
cKEY_BACKSPACE = #const KEY_BACKSPACE
cKEY_F0 = #const KEY_F0
-- cKEY_F(n) = #const KEY_F(n)
cKEY_DL = #const KEY_DL
cKEY_IL = #const KEY_IL
cKEY_DC = #const KEY_DC
cKEY_IC = #const KEY_IC
cKEY_EIC = #const KEY_EIC
cKEY_CLEAR = #const KEY_CLEAR
cKEY_EOS = #const KEY_EOS
cKEY_EOL = #const KEY_EOL
cKEY_SF = #const KEY_SF
cKEY_SR = #const KEY_SR
cKEY_NPAGE = #const KEY_NPAGE
cKEY_PPAGE = #const KEY_PPAGE
cKEY_STAB = #const KEY_STAB
cKEY_CTAB = #const KEY_CTAB
cKEY_CATAB = #const KEY_CATAB
cKEY_ENTER = #const KEY_ENTER
cKEY_SRESET = #const KEY_SRESET
cKEY_RESET = #const KEY_RESET
cKEY_PRINT = #const KEY_PRINT
cKEY_LL = #const KEY_LL
cKEY_A1 = #const KEY_A1
cKEY_A3 = #const KEY_A3
cKEY_B2 = #const KEY_B2
cKEY_C1 = #const KEY_C1
cKEY_C3 = #const KEY_C3

cKEY_BTAB, cKEY_BEG, cKEY_CANCEL, cKEY_CLOSE, cKEY_COMMAND, cKEY_COPY, cKEY_CREATE, cKEY_END, cKEY_EXIT, cKEY_FIND, cKEY_HELP, cKEY_MARK, cKEY_MESSAGE, cKEY_MOVE, cKEY_NEXT, cKEY_OPEN, cKEY_OPTIONS, cKEY_PREVIOUS, cKEY_REDO, cKEY_REFERENCE, cKEY_REFRESH, cKEY_REPLACE, cKEY_RESIZE, cKEY_RESTART, cKEY_RESUME, cKEY_SAVE, cKEY_SBEG, cKEY_SCANCEL, cKEY_SCOMMAND, cKEY_SCOPY, cKEY_SCREATE, cKEY_SDC, cKEY_SDL, cKEY_SELECT, cKEY_SEND, cKEY_SEOL, cKEY_SEXIT, cKEY_SFIND, cKEY_SHELP, cKEY_SHOME, cKEY_SIC, cKEY_SLEFT, cKEY_SMESSAGE, cKEY_SMOVE, cKEY_SNEXT, cKEY_SOPTIONS, cKEY_SPREVIOUS, cKEY_SPRINT, cKEY_SREDO, cKEY_SREPLACE, cKEY_SRIGHT, cKEY_SRSUME, cKEY_SSAVE, cKEY_SSUSPEND, cKEY_SUNDO, cKEY_SUSPEND, cKEY_UNDO  :: CInt
cKEY_BTAB = #const KEY_BTAB
cKEY_BEG = #const KEY_BEG
cKEY_CANCEL = #const KEY_CANCEL
cKEY_CLOSE = #const KEY_CLOSE
cKEY_COMMAND = #const KEY_COMMAND
cKEY_COPY = #const KEY_COPY
cKEY_CREATE = #const KEY_CREATE
cKEY_END = #const KEY_END
cKEY_EXIT = #const KEY_EXIT
cKEY_FIND = #const KEY_FIND
cKEY_HELP = #const KEY_HELP
cKEY_MARK = #const KEY_MARK
cKEY_MESSAGE = #const KEY_MESSAGE
cKEY_MOVE = #const KEY_MOVE
cKEY_NEXT = #const KEY_NEXT
cKEY_OPEN = #const KEY_OPEN
cKEY_OPTIONS = #const KEY_OPTIONS
cKEY_PREVIOUS = #const KEY_PREVIOUS
cKEY_REDO = #const KEY_REDO
cKEY_REFERENCE = #const KEY_REFERENCE
-- KEY_REFRESH not standard:
cKEY_REFRESH = #const KEY_REFRESH
cKEY_REPLACE = #const KEY_REPLACE
cKEY_RESIZE = #const KEY_RESIZE
cKEY_RESTART = #const KEY_RESTART
cKEY_RESUME = #const KEY_RESUME
cKEY_SAVE = #const KEY_SAVE
cKEY_SBEG = #const KEY_SBEG
cKEY_SCANCEL = #const KEY_SCANCEL
cKEY_SCOMMAND = #const KEY_SCOMMAND
cKEY_SCOPY = #const KEY_SCOPY
cKEY_SCREATE = #const KEY_SCREATE
cKEY_SDC = #const KEY_SDC
cKEY_SDL = #const KEY_SDL
cKEY_SELECT = #const KEY_SELECT
cKEY_SEND = #const KEY_SEND
cKEY_SEOL = #const KEY_SEOL
cKEY_SEXIT = #const KEY_SEXIT
cKEY_SFIND = #const KEY_SFIND
cKEY_SHELP = #const KEY_SHELP
cKEY_SHOME = #const KEY_SHOME
cKEY_SIC = #const KEY_SIC
cKEY_SLEFT = #const KEY_SLEFT
cKEY_SMESSAGE = #const KEY_SMESSAGE
cKEY_SMOVE = #const KEY_SMOVE
cKEY_SNEXT = #const KEY_SNEXT
cKEY_SOPTIONS = #const KEY_SOPTIONS
cKEY_SPREVIOUS = #const KEY_SPREVIOUS
cKEY_SPRINT = #const KEY_SPRINT
cKEY_SREDO = #const KEY_SREDO
cKEY_SREPLACE = #const KEY_SREPLACE
cKEY_SRIGHT = #const KEY_SRIGHT
cKEY_SRSUME = #const KEY_SRSUME
cKEY_SSAVE = #const KEY_SSAVE
cKEY_SSUSPEND = #const KEY_SSUSPEND
cKEY_SUNDO = #const KEY_SUNDO
cKEY_SUSPEND = #const KEY_SUSPEND
cKEY_UNDO = #const KEY_UNDO

cTRUE :: CBool
cTRUE = #const TRUE
cFALSE :: CBool
cFALSE = #const FALSE

cBLACK, cBLUE, cGREEN, cCYAN, cRED, cMAGENTA, cYELLOW, cWHITE :: CShort
cBLACK = #const COLOR_BLACK
cBLUE = #const COLOR_BLUE
cGREEN = #const COLOR_GREEN
cCYAN = #const COLOR_CYAN
cRED = #const COLOR_RED
cMAGENTA = #const COLOR_MAGENTA
cYELLOW = #const COLOR_YELLOW
cWHITE = #const COLOR_WHITE

aUNDERLINE :: CInt
aUNDERLINE = #const A_UNDERLINE
aNORMAL :: CInt
aNORMAL = #const A_NORMAL
aBOLD :: CInt
aBOLD = #const A_BOLD

class IsChType a where
    to_ChType :: a -> ChType

instance IsChType Char where
    to_ChType c = fromIntegral (ord c)

instance IsChType ChType where
    to_ChType c = c

class IsChStr a where
    to_ChStr :: a -> IO (Ptr ChType)
    used_ChStr :: a -> Ptr ChType -> IO ()
    used_ChStr _ p = free p

instance IsChStr (Ptr ChType) where
    to_ChStr s = return s
    used_ChStr _ _ = return ()

instance IsChStr [ChType] where
    to_ChStr cs = newArray0 0 cs

class IsCString a where
    to_CString :: a -> IO CString
    used_CString :: a -> CString -> IO ()
    used_CString _ cs = free cs

instance IsCString CString where
    to_CString s = return s
    used_CString _ _ = return ()

instance IsCString String where
    to_CString s = newCString s

-- Constants
$(
    let fs = map ("acs_" ++)
           $ [which ++ "corner" | which <- ["ul", "ll", "ur", "lr"]]
          ++ [which ++ "tee" | which <- ["r", "l", "b", "t"]]
          ++ ["hline", "vline", "plus", "s1", "s9", "diamond", "ckboard",
              "degree", "plminus", "bullet"]
          ++ [which ++ "arrow" | which <- ["l", "r", "d", "u"]]
          ++ ["board", "lantern", "block"]
        imp f = ForeignD (
                    ImportF CCall Safe ("wrap.h " ++ f) (mkName f)
                    (AppT (ConT (mkName "IO")) (ConT (mkName "ChType")))
                )
        h f = "chtype " ++ f ++ "(void);\n"
        c f = "chtype " ++ f ++ "(void) {\n"
           ++ "    return " ++ map toUpper f ++ ";\n"
           ++ "}\n"
        c_header = "#include \"wrap.h\"\n"
                ++ "void w_getmaxyx(WINDOW *w, int *y, int *x) {\n"
                ++ "    getmaxyx(w, *y, *x);\n"
                ++ "}\n"
                ++ "void w_getyx(WINDOW *w, int *y, int *x) {\n"
                ++ "    getyx(w, *y, *x);\n"
                ++ "}\n"
        h_header = "#include <curses.h>\n"
                ++ "void w_getmaxyx(WINDOW *w, int *y, int *x);\n"
                ++ "void w_getyx(WINDOW *w, int *y, int *x);\n"
    in do runIO $ do writeFile "wrap.h" $ h_header ++ concatMap h fs
                     writeFile "wrap.c" $ c_header ++ concatMap c fs
                     -- XXX hack: cabal does the compiling for now
                     -- system "gcc -Wall -pedantic -c wrap.c -o wrap.o"
          return $ map imp fs
 )

$( do cont <- runIO $ readFile "curses_header"
      {-
      ds <- gen_decs cont
      runIO $ putStrLn $ pprint ds
      return ds
      -}
      gen_decs cont
 )

foreign import ccall safe "static haskell-curses.h &stdscr" c_stdscr
            :: Ptr PWindow
stdscr :: MonadIO io => io PWindow
stdscr = liftIO $ peek c_stdscr

foreign import ccall safe "static haskell-curses.h &curscr" c_curscr
            :: Ptr PWindow
curscr :: MonadIO io => io PWindow
curscr = liftIO $ peek c_curscr

-----
foreign import ccall safe "static wrap.h w_getmaxyx" w_getmaxyx
    :: PWindow -> Ptr CInt -> Ptr CInt -> IO ()
getmaxyx :: MonadIO io => PWindow -> io (CInt, CInt)
getmaxyx w = liftIO $ wrap_getxys w_getmaxyx w

foreign import ccall safe "static wrap.h w_getyx" w_getyx
    :: PWindow -> Ptr CInt -> Ptr CInt -> IO ()
getyx :: MonadIO io => PWindow -> io (CInt, CInt)
getyx w = wrap_getxys w_getyx w

wrap_getxys :: MonadIO io
            => (PWindow -> Ptr CInt -> Ptr CInt -> IO ())
            -> PWindow -> io (CInt, CInt)
wrap_getxys f w = liftIO $
                  alloca $ \py ->
                  alloca $ \px ->
                  do f w py px
                     y <- peek py
                     x <- peek px
                     return (y, x)
-----

{-
XXX

errI :: MonadIO io => io CInt -> io ()
errI f = do r <- f
            if r == cERR then do endwin
                                 error "curses returned an error"
                         else return ()

errP :: MonadIO io => io (Ptr a) -> io ()
errP f = do p <- f
            if p == nullPtr then do endwin
                                    error "curses returned an error"
                            else return ()
-}

